/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment media rendering library
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Loïc Molinari <loic@fluendo.com>
 */

#ifndef __PGM_EVENTS_H__
#define __PGM_EVENTS_H__

/* pgmevents.h and pgmviewport.h include eachother */
typedef struct _PgmEventAny       PgmEventAny;
typedef struct _PgmEventMotion    PgmEventMotion;
typedef struct _PgmEventButton    PgmEventButton;
typedef struct _PgmEventScroll    PgmEventScroll;
typedef struct _PgmEventKey       PgmEventKey;
typedef struct _PgmEventExpose    PgmEventExpose;
typedef struct _PgmEventConfigure PgmEventConfigure;
typedef struct _PgmEventDelete    PgmEventDelete;
typedef union  _PgmEvent          PgmEvent;

#include <glib-object.h>

G_BEGIN_DECLS

#define PGM_TYPE_EVENT (pgm_event_get_type ())

/**
 * PgmEventType:
 * @PGM_NOTHING: a special code to indicate a null event.
 * @PGM_MOTION_NOTIFY: the pointer has entered the window.
 * @PGM_BUTTON_PRESS: a mouse button has been pressed.
 * @PGM_DOUBLE_BUTTON_PRESS: a mouse button has been clicked 2 times in a short
 * period of time. Note that each click also generates a %PGM_BUTTON_PRESS
 * event.
 * @PGM_TRIPLE_BUTTON_PRESS: a mouse button has been clicked 3 times in a short
 * period of time. Note that each click also generates a %PGM_BUTTON_PRESS
 * event.
 * @PGM_BUTTON_RELEASE: a mouse button has been released.
 * @PGM_KEY_PRESS: a key has been pressed.
 * @PGM_KEY_RELEASE: a key has been released.
 * @PGM_EXPOSE: the window has become visible and needs to be redrawn.
 * @PGM_CONFIGURE: the size or the position of the viewport has changed.
 * @PGM_SCROLL: the scroll wheel was turned.
 * @PGM_DELETE: the window manager has requested that the toplevel window be
 * destroyed, usually when the user clicks on a special icon in the title bar.
 *
 * Specifies the type of the event.
 */
typedef enum {
  PGM_NOTHING             = -1,
  PGM_MOTION_NOTIFY       = 0,
  PGM_BUTTON_PRESS        = 1,
  PGM_DOUBLE_BUTTON_PRESS = 2,
  PGM_TRIPLE_BUTTON_PRESS = 3,
  PGM_BUTTON_RELEASE      = 4,
  PGM_KEY_PRESS           = 5,
  PGM_KEY_RELEASE         = 6,
  PGM_EXPOSE              = 7,
  PGM_CONFIGURE           = 8,
  PGM_SCROLL              = 9,
  PGM_DELETE              = 10
} PgmEventType;

/**
 * PgmModifierType:
 * @PGM_SHIFT_MASK: the Shift key.
 * @PGM_CAPSLOCK_MASK: the Caps Lock.
 * @PGM_CONTROL_MASK: the Control key.
 * @PGM_ALT_MASK: the Alt key.
 * @PGM_NUMLOCK_MASK: the Num Lock.
 *
 * A set of bit-flags to indicate the state of modifier keys. Typical modifier
 * keys are Shift, Control, Alt and CapsLock.
 */
typedef enum {
  PGM_SHIFT_MASK    = (1 << 0),
  PGM_CAPSLOCK_MASK = (1 << 1),
  PGM_CONTROL_MASK  = (1 << 2),
  PGM_ALT_MASK      = (1 << 3),
  PGM_NUMLOCK_MASK  = (1 << 4)
} PgmModifierType;

/**
 * PgmButtonType:
 * @PGM_BUTTON_LEFT: the left mouse button.
 * @PGM_BUTTON_MIDDLE: the middle mouse button.
 * @PGM_BUTTON_RIGHT: the right mouse button.
 *
 * The mouse button type.
 */
typedef enum {
  PGM_BUTTON_LEFT   = (1 << 0),
  PGM_BUTTON_MIDDLE = (1 << 1),
  PGM_BUTTON_RIGHT  = (1 << 2)
} PgmButtonType;

/**
 * PgmScrollDirection:
 * @PGM_SCROLL_UP: the scrolling in the upward direction.
 * @PGM_SCROLL_DOWN: the scrolling in the downward direction.
 *
 * The mouse wheel scrolling directions.
 */
typedef enum {
  PGM_SCROLL_UP,
  PGM_SCROLL_DOWN
} PgmScrollDirection;

/**
 * PgmEventAny:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 *
 * Contains the fields which are common to all event structs. Any event
 * pointer can safely be cast to a pointer to a #PgmEventAny to access these
 * fields.
 */
struct _PgmEventAny {
  PgmEventType type;
  guint8       source;
};

/**
 * PgmEventMotion:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 * @time: the time of the event in milliseconds.
 * @x: the x coordinate of the pointer relative to the window in canvas
 * coordinate.
 * @y: the y coordinate of the pointer relative to the window in canvas
 * coordinate.
 *
 * Generated when the pointer moves.
 */
struct _PgmEventMotion {
  PgmEventType type;
  guint8       source;
  guint32      time;
  gfloat       x, y;
};

/**
 * PgmEventButton:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 * @time: the time of the event in milliseconds.
 * @x: the x coordinate of the pointer relative to the window in canvas
 * coordinate.
 * @y: the y coordinate of the pointer relative to the window in canvas
 * coordinate.
 * @button: the button which was pressed or released.
 *
 * Used for button press and button release events.
 */
struct _PgmEventButton {
  PgmEventType  type;
  guint8        source;
  guint32       time;
  gfloat        x, y;
  PgmButtonType button;
};

/**
 * PgmEventScroll:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 * @time: the time of the event in milliseconds.
 * @x: the x coordinate of the pointer relative to the window in canvas
 * coordinate.
 * @y: the y coordinate of the pointer relative to the window in canvas
 * coordinate.
 * @direction: the scroll wheel direction.
 *
 * Generated when the mouse wheel is turned.
 */
struct _PgmEventScroll {
  PgmEventType       type;
  guint8             source;
  guint32            time;
  gfloat             x, y;
  PgmScrollDirection direction;
};

/**
 * PgmEventKey:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 * @time: the time of the event in milliseconds.
 * @modifier: A bit-mask representing the state of the modifier keys (e.g.
 * Control, Shift and Alt), see #PgmModifierType.
 * @keyval: the key that was pressed or released. See the pgm/pgmkeysyms.h
 * header file for a complete list of Pigment key codes.
 * @hardware_keycode: the raw code of the key that was pressed or released.
 *
 * Describes a key press or key release event.
 */
struct _PgmEventKey {
  PgmEventType type;
  guint8       source;
  guint32      time;
  guint        modifier;
  guint        keyval;
  guint16      hardware_keycode;
};

/**
 * PgmEventExpose:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 *
 * Generated when a window becomes visible and needs to be redrawn.
 */
struct _PgmEventExpose {
  PgmEventType type;
  guint8       source;
};

/**
 * PgmEventConfigure:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 * @x: the new x coordinate of the window.
 * @y: the new y coordinate of the window.
 * @width: the new width of the viewport.
 * @height: the new height of the viewport.
 *
 * Generated when a viewport size or position has changed.
 */
struct _PgmEventConfigure {
  PgmEventType type;
  guint8       source;
  gint         x, y;
  gint         width, height;
};

/**
 * PgmEventDelete:
 * @type: the type of the event.
 * @source: the source field that can be filled by applications to indicate the
 * source of the event (mouse, pen, remote control, etc).
 * @time: the time of the event in milliseconds.
 *
 * Generated when the user requested that the viewport be destroyed.
 */
struct _PgmEventDelete {
  PgmEventType type;
  guint8       source;
  guint32      time;
};

/**
 * PgmEvent:
 *
 * <para>
 * The #PgmEvent struct contains a union of all of the event structs,
 * and allows access to the data fields in a number of ways.
 * </para>
 * <para>
 * The event type is always the first field in all of the event structs, and
 * can always be accessed with the following code, no matter what type of event
 * it is:
 * <informalexample>
 * <programlisting language="c">
 *   PgmEvent *event;  
 *   PgmEventType type;
 *   type = event->type;
 * </programlisting>
 * </informalexample>
 * </para>
 * <para>
 * To access other fields of the event structs, the pointer to the event can be
 * cast to the appropriate event struct pointer, or the union member name can be
 * used. For example if the event type is %PGM_BUTTON_PRESS then the x coordinate
 * of the button press can be accessed with:
 * <informalexample>
 * <programlisting language="c">
 *   PgmEvent *event;  
 *   gfloat x;
 *   x = ((PgmEventButton*) event)->x;
 * </programlisting>
 * </informalexample>
 * or with:
 * <informalexample>
 * <programlisting language="c">
 *   PgmEvent *event;  
 *   gfloat x;
 *   x = event->button.x;
 * </programlisting>
 * </informalexample>
 * </para>
 */
union _PgmEvent {
  PgmEventType      type;
  PgmEventAny       any;
  PgmEventMotion    motion;
  PgmEventButton    button;
  PgmEventScroll    scroll;
  PgmEventKey       key;
  PgmEventExpose    expose;
  PgmEventConfigure configure;
  PgmEventDelete    delete;
};

GType     pgm_event_get_type    (void);

PgmEvent *pgm_event_new         (PgmEventType type);
PgmEvent *pgm_event_copy        (PgmEvent *event);
void      pgm_event_free        (PgmEvent *event);

guint32   pgm_keyval_to_unicode (guint keyval) G_GNUC_CONST;

G_END_DECLS

#endif /* __PGM_EVENTS_H__ */
